--- drweb-configure session module
-- @copyright 2019 Dr.Web Ltd.

--- Initialize augeas object, and load pam config
-- @return augeas object
local function init()
  local augeas = require("augeas")
  local aug = augeas.init({ noautoload = true, save_backup = false, no_load = true, no_modl_autoload = true })

  aug:set("/augeas/load/Pam/lens", "Pam.lns")
  aug:set("/augeas/load/Pam/incl", "/etc/pam.d/*")
  aug:set("/augeas/load/Simplelines/lens", "Simplelines.lns")
  aug:set("/augeas/load/Simplelines/incl", "/etc/parsec/*")

  aug:load()

  if aug:error() > 0 then
    io.stderr:write(aug:error_message(), "\n")
    aug:close()
    return nil
  end

  return aug
end

--- Print usage information
local function usage()
  io.write(string.format("Usage: %s %s { enable | disable }\n", arg[-1], arg[0]))
end

--- Check if a path exists
-- @param path path to check
-- @return true or false
local function exists(path)
  return os.rename(path, path) or false
end

--- Get PAM module directory path
-- @return path
local function get_pam_module_path()
  local output = io.popen("/sbin/ldconfig -p"):read("a")
  local libpam_path = output:match("%S*/libpam.so%S*")
  if not libpam_path then
    error("Cannot find PAM module directory", 0)
  end
  local pam_module_path = libpam_path:gsub("libpam.so.*", "security/")
  return pam_module_path
end

--- Get one path matching pattern
-- @param pattern pattern string
-- @return path
local function do_glob(pattern)
  local found = require("posix.glob").glob(pattern, 0)
  if found then
    for _, path in pairs(found) do
      return path
    end
  end
  error("Cannot find " .. pattern, 0)
end

--- Disable drweb-session
-- @param aug augeas object
-- @return true or false
local function disable(aug)
  aug:rm("/files/etc/pam.d/*/*[./module = 'pam_drweb_session.so']")

  if exists("/etc/parsec") then
    aug:rm("/files/etc/parsec/privsock.conf/*['/opt/drweb.com/bin/drweb-configd']")
    aug:rm("/files/etc/parsec/privsock.conf/*['/opt/drweb.com/bin/drweb-configd.real']")
  end

  -- remove symlink pam_drweb_session.so
  local ok, err = pcall(function()
    local linkpath = get_pam_module_path() .. "pam_drweb_session.so"
    local ok, err, ec = os.remove(linkpath)
    if not ok and ec ~= require("posix").ENOENT then
      error(err, 0)
    end
  end)
  if not ok then
    io.stderr:write("Error: " .. err .. "\n")
  end

  return ok
end

--- Enable drweb-session
-- @param aug augeas object
-- @return true or false
local function enable(aug)
  -- first ensure there are no pam_drweb_session.so records
  if not disable(aug) then
    return false
  end

  if exists("/etc/parsec") then
    -- configuration privsok file for v1.6
    aug:set("/files/etc/parsec/privsock.conf/01", '/opt/drweb.com/bin/drweb-configd')
    aug:set("/files/etc/parsec/privsock.conf/02", '/opt/drweb.com/bin/drweb-configd.real')
  end

  -- create symlink pam_drweb_session.so
  local ok, err = pcall(function()
    local target = do_glob("/opt/drweb.com/lib/*/pam/pam_drweb_session.so")
    local linkpath = get_pam_module_path() .. "pam_drweb_session.so"
    local ret, err = require("posix.unistd").link(target, linkpath, true)
    if not ret then
      error(err, 0)
    end
  end)
  if not ok then
    io.stderr:write("Error: " .. err .. "\n")
    return false
  end

  -- configuration files where the pam_parsec_mac.so module is mentioned
  local confs = aug:match("/files/etc/pam.d/*[descendant::*[type = 'session'][module = 'pam_parsec_mac.so' or module = 'pam_namespace.so']]")

  for i, node in ipairs(confs) do
    aug:insert(node .. "/*[./type = 'session'][1]", "01", true)
    aug:set(node .. "/01/type", "session")
    aug:set(node .. "/01/control", "optional")
    aug:set(node .. "/01/module", "pam_drweb_session.so")
    aug:set(node .. "/01/argument", "type=close")

    aug:insert(node .. "/*[./type = 'session'][last()]", "02", false)
    aug:set(node .. "/02/type", "session")
    aug:set(node .. "/02/control", "optional")
    aug:set(node .. "/02/module", "pam_drweb_session.so")
    aug:set(node .. "/02/argument", "type=open")
  end

  return true
end

function session_main()
  local status = 0

  if arg[1] == nil or arg[2] ~= nil then
    usage()
    return 1
  end

  local aug = init()
  if not aug then
    io.stderr:write("Error: Cannot initialize augeas\n")
    return 2
  end

  if arg[1] == "enable" then
    if not enable(aug) then
      return 3
    end
  elseif arg[1] == "disable" then
    if not disable(aug) then
      status = 3
    end
  else
    usage()
    return 1
  end

  if aug:save() ~= 0 then
    io.stderr:write("Error: Cannot save changes. Make sure that you run this command with Superuser privileges\n")
    aug:close()
    return 5
  end

  return status
end

function session_help()
  usage()
end
